#ifndef _TYPES_HH_
#define _TYPES_HH_

class PrimitiveType;
class ContainerType;
class TupleType;
class CollectionType;
class IndexedCollectionType;
class ObjectType;
class StructType;
class ListType;
class ArrayType;
class MethodType;
class SetType;
class BagType;
class RefType;

class Descriptor;

class Type {
	friend	ostream &operator<<(ostream &s, Type &type);
protected:
	auto_string	_name;
	bool	_saved;	

public: /* protected: */
	Type(char *name = 0) : _name(name) { }

public:
	bool	isNamed() { return _name.str()[0] != 0; }
	virtual char	*name();
	void	setName(char *name) {
		_name.set(name);
	}

	bool	isSaved() { return _saved; } 

	/* Downcasts */ 
	virtual	PrimitiveType	*isPrimitive() { return 0; }

	virtual	ContainerType	*isContainer() { return 0; }
	virtual	TupleType	*isTuple() { return 0; }
	virtual	ObjectType	*isObject() { return 0; }
	virtual StructType	*isStruct() { return 0; }

	virtual	MethodType	*isMethod() { return 0; }

	virtual	CollectionType	*isCollection() { return 0; }
	virtual IndexedCollectionType	*isIndexedCollection() { return 0;}
	virtual	ArrayType	*isArray() { return 0; }
	virtual	ListType	*isList()  { return 0; }
	virtual SetType		*isSet() { return 0; }
	virtual BagType		*isBag() { return 0; }

	virtual RefType	*isRef() { return 0; }

	/* psuedo-Downcasts */
	virtual	int	isInteger() { return 0; }
	virtual int	isBoolean() { return 0; }

	/* arithmetic and object compatabilities */
	virtual Type	*canCompareWith(Type *it) { return 0; }
	virtual	Type	*canDoMathWith(Type *it) { return 0; }
	virtual Type	*canDoMath() { return 0; }
	virtual	Type	*lcs(Type *it) { return 0; }

	virtual ostream	&print(ostream &s) {
		return s << "<Type " << Type::name() << ">";
	}

	/* methods for realizing types */
	virtual	void	commit() {};
	virtual	int	size() { return 0; }
	virtual	int	inline_size() { return size(); }
	virtual int	align() { return inline_size(); }

	/* Default save action is to not do anything */
	virtual	void	save(ostream &) { };
};

class ErrorType : public Type {
public: /* protected: */
	ErrorType(char *name) : Type(name) {}
public:
};

extern ErrorType TypeError;


class PrimitiveType : public Type {
public:
#include "primitive_types.h"
private:
	enum PrimitiveTypes	_primitive;
public: /*protected:*/
	PrimitiveType(char *name, PrimitiveTypes prim) : Type(name) {
		_primitive = prim;
		_saved = true;	/* never written out */
	}
			
public:
	virtual	PrimitiveType	*isPrimitive() { return this; }
	enum PrimitiveTypes	primitive() { return _primitive; }
	Type	*lcs(Type *it);
	Type	*canCompareWith(Type *it);
	Type	*canDoMathWith(Type *it);
	Type	*canDoMath();
	int	isInteger() { return _primitive == pt_integer; }
	int	isBoolean() { return _primitive == pt_boolean; }
	// ostream	&print(ostream &s);

	int	size();
	int	inline_size();

	void	extract(Descriptor &d, int offset, void *); 
};

extern PrimitiveType TypeVoid;

/*
   Base type for types containing named members
   This is really a reusable interface class for container objects.
 */


class ContainerType_i;

class ContainerType : public Type {
	friend class ContainerType_i;
protected: /* private: */
	SymbolTable	members;
	int	_size;
	int	_align;
	bool	commit_started;
	struct	TableEntry {
		Type	*type;
		int	offset;
		TableEntry(Type *t, int _offset = 0) {
			offset = _offset;
			type = t;
		}
	};
public: /* protected: */
	ContainerType(char *name) : Type(name) {
		_size = 0;
		commit_started = false;
	}
	/* This is only visible for tuples in real life */
	int	add(char *name, Type *type) {
		Symbol *s = members.lookup(name);
		if (s)
			return 1;
		TableEntry *entry = new TableEntry(type);
		
		members.enter(name, (char *)entry);
		return 0;
	}
	ostream	&_print(ostream &s, char *typename);
	void	_commit();
	TableEntry	*_member(char *name) {
		Symbol	*s = members.lookup(name);
		return s ? (TableEntry *)(s->data()) : (TableEntry *)0;
	}
public:
	virtual Type	*member(char *name) {
		TableEntry *e = _member(name);
		return e ? e->type : 0;
	}
	virtual Type	*member(char *name, int &offset) {
		TableEntry *e = _member(name);
		if (e) { 
			offset = e->offset;
			return e->type;
		}
		return 0;
	}
	virtual	ostream	&print(ostream &s) = 0;
	virtual	ContainerType	*isContainer() { return this; }
	virtual	Type	*canCompareWith(Type *of);

	void	commit();
	int	size() { return _size; }
	int	align();

	void	save(ostream &);

	virtual	ContainerType_i	*iterator();	// over members
};


class ObjectType : public ContainerType {
private:
	ObjectType	*_superclass;
	void classlist(ObjectType **list, int &where, int remaining); 
public: /*protected:*/
	ObjectType(char *name, ObjectType *superclass = 0) : ContainerType(name) {
		_superclass = superclass;
			
	}
	TableEntry *_member(char *name) {
		TableEntry	*e;
		e = ContainerType::_member(name);
		if (!e && _superclass)
			e = _superclass->_member(name);
		return e;
	}
public:
	Type	*member(char *name) {
		TableEntry *e = _member(name);
		return e ? e->type : 0 ;
	}
	Type	*member(char *name, int &offset) {
		TableEntry *e = _member(name);
		if (e) {
			offset = e->offset;
			return e->type;
		}
		return 0;
	}
	int	isa(ObjectType *it);
	Type	*lcs(Type *it);
	virtual	ObjectType	*isObject() { return this; }
	ostream	&print(ostream &s) { return ContainerType::_print(s, 0);}

	/* XXX need something better for multiple inheritance */
	ObjectType	*superclass() { return _superclass; }
	void	setSuperclass(ObjectType *superclass) {
		_superclass = superclass;
	}

	int	size() {
		return  _size + (_superclass ? _superclass->size() : 0);
	}
	int	inline_size() { return sizeof(void *); }
	int	align() { return inline_size(); }
	void	commit();
};


class StructType : public ContainerType {
private:
public: /* protected: */
	StructType(char *name = 0) : ContainerType(name) {};
public:
	virtual	StructType	*isStruct() { return this; }
	ostream	&print(ostream &s) { return ContainerType::_print(s, "Struct");}
};


class TupleType : public ContainerType {
private:
public: /* protected: */
	TupleType(char *name = 0) : ContainerType(name) {};
public:
	void concat(TupleType &from);
	char	*name();
	Type	*lcs(Type *it);
	virtual	TupleType	*isTuple() { return this; }
	ostream	&print(ostream &s) { return ContainerType::_print(s, "Tuple");}
};


class MethodType : public Type {
	friend class MethodType_i;
private:
	Type	*_returns;
	SymbolTable	args;

public: /*  protected: */
	MethodType(char *name, Type *returns = &TypeVoid) : Type(name) {
		_returns = returns;
	}
	MethodType(Type *returns = &TypeVoid) : Type(0) {
		_returns = returns;
	}
	int	addArg(char *name, Type *type) {
		Symbol *s = args.lookup(name);
		if (s)
			return 1;
		args.enter(name, (char *)type, 0);
		return 0;
	}
	void	setReturns(Type *type) {
		_returns = type;
	}

public:
	Type	*returns() { return _returns; }
	Type	*argument(char *name) {
		Symbol	*s = args.lookup(name);
		return s ? (Type *)(s->data()) : (Type *)0;
	}
	virtual	MethodType	*isMethod() { return this; }

	MethodType_i	*iterator();	// over args

	ostream	&print(ostream &s);
	void	save(ostream &);
};


/*
   XXXX
   
   Reference share the interface that Collections use,
   but a reference isn't a collection, so it shouldn't be
   part of that class tree.
 */  

class RefType : public Type {
private:
	Type	*_of;
public: /* protected: */
	RefType(char *name, Type *of) : Type(name) {
		_of = of;
	}
	RefType(Type *of) : Type((char *)0) {
		_of = of;
	}
	void setOf(Type *of) { _of = of; }
public:
	Type	*of() { return _of; }
	virtual	RefType	*isRef() { return this; }
	Type	*canCompareWith(Type *of);
	Type	*rebuild(Type *of);

	Type	*lcs(Type *it);

	// Type	*operator->() { return _of; }
	Type *operator*() { return _of; }
	
	ostream	&print(ostream &s);
};

#if 0
class NilRefType : public RefType {
};
#endif

/*
   Collections of similar types

   As with the container, this class is a reusable interface class
 */

class CollectionAccessor;

class CollectionType : public Type {
private:
	Type	*_of;
public: /* protected: */
	CollectionType(char *name, Type *of) : Type(name) {
		_of = of;
	}
	CollectionType(Type *of) : Type((char *)0) {
		_of = of;
	}
	void setOf(Type *of) { _of = of; }
	ostream	&_print(ostream &s, char *typename);
public:
	virtual	Type	*rebuild(Type *of) = 0;
	Type	*of() { return _of; }
	virtual	CollectionType	*isCollection() { return this; }

	virtual Type	*lcs(Type *it) = 0;
	Type	*_lcs(CollectionType *it);
	
	virtual ostream	&print(ostream &s) = 0;

	int	inline_size() { return sizeof(int) + sizeof(void *); }
	int	align() { return sizeof(int); }
	void	commit() { _of->commit(); }

	CollectionAccessor	*accessor(Descriptor &d, int offset = 0);
};


#if 0
class RefType : public CollectionType {
public:
	RefType(char *name, Type *of) : CollectionType(name, of) {}
	RefType(Type *of) : CollectionType(of) {}
};
#endif

class BagType : public CollectionType {
public:
	virtual BagType		*isBag() { return this; }
	BagType(char *name, Type *of) : CollectionType(name, of) {}
	BagType(Type *of) : CollectionType(of) {}
	virtual	Type	*rebuild(Type *of) { return new BagType(of); }
	virtual Type	*lcs(Type *it);
	ostream	&print(ostream &s) { return CollectionType::_print(s, "Bag");}
};

class SetType : public BagType {
public:
	virtual SetType		*isSet() { return this; }
	SetType(char *name, Type *of) : BagType(name, of) {}
	SetType(Type *of) : BagType(of) {}
	virtual	Type	*rebuild(Type *of) { return new SetType(of); }
	Type	*lcs(Type *it);
	ostream	&print(ostream &s) { return CollectionType::_print(s, "Set");}
};

class IndexedCollectionType : public CollectionType {
	int	_cardinality;
public:
	IndexedCollectionType(char *name, Type *of) : CollectionType(name,of) {
		_cardinality = 0;
	}
	IndexedCollectionType(Type *of) : CollectionType(of) {
		_cardinality = 0;
	}
	virtual IndexedCollectionType	*isIndexedCollection() { return this;}
	int	cardinality() { return _cardinality; }
	void setCardinality(int newcard) { _cardinality = newcard; }
	virtual	Type	*rebuild(Type *of) = 0;
	virtual	Type	*lcs(Type *it) = 0;
	virtual	ostream	&print(ostream &s) = 0;
};

class ListType : public IndexedCollectionType {
public:
	virtual	ListType	*isList()  { return this; }
	ListType(char *name, Type *of) : IndexedCollectionType(name, of) {}
	ListType(Type *of) : IndexedCollectionType(of) {}
	virtual	Type	*rebuild(Type *of) { return new ListType(of); }
	Type	*lcs(Type *it);
	ostream	&print(ostream &s) { return CollectionType::_print(s, "List");}
};

class ArrayType : public IndexedCollectionType {
public:
	virtual	ArrayType	*isArray() { return this; }
	ArrayType(char *name, Type *of) : IndexedCollectionType(name, of) {}
	ArrayType(Type *of) : IndexedCollectionType(of) {}
	virtual	Type	*rebuild(Type *of) { return new ArrayType(of); }
	Type	*lcs(Type *it);
	ostream	&print(ostream &s) { return CollectionType::_print(s, "Array");}
};


extern ObjectType TypeNil; 

class ContainerType_i {
	SymbolTable_i	_member;
public:
	ContainerType_i(ContainerType &it) : _member(it.members) {}
	bool	next(char *&name, Type *&type, int &offset);
};

class MethodType_i {
	SymbolTable_i	_args;
public:
	MethodType_i(MethodType &it) : _args(it.args) { }
	bool	next(char *&name, Type *&type);
};

class CollectionAccessor {
public:
	virtual	int	count() = 0;
	virtual	Type	&of() = 0;
	virtual	bool	at(int i, Descriptor &d) = 0;
	virtual	bool	next(Descriptor &d) = 0 ;
	virtual	bool	more() = 0;
};

#endif /* _TYPES_HH_ */
