

/*
 * TCL for reading node description files and 
 * filling in the node information.
 */
#include <stdlib.h>
#include <stddef.h>
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#include "symbol.hh"
#include "types.hh"
#include "typedb.hh"
#include "oql_context.hh"	

#include "local_tcl.hh"

typedef	enum {RCOK=0, RC_GENERIC=1}	rc_t;

class tclContext {
	enum	{DEPTH=20};
	Type	*_stack[DEPTH];
	int	_tos;
public:
	Tcl_Interp	*tcl;
	oqlDatabase	&db;
	tclContext(oqlDatabase &the_db) : db(the_db) {
		_tos = 0;
	}

	Type	*&top() { return _stack[_tos]; }
	void	push(Type *it) { _stack[++_tos] = it; }
	Type	*pop() { return _stack[_tos--]; }

	void	add_type(char *typename, Type *type = 0);
};


static	int	_load_type(istream &from, tclContext *context);
extern	rc_t	load_type(istream &from, oqlDatabase &db);


static define(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	tclContext *context = (tclContext *)_context;
	Type	*it;
	char	*type_name;
	int	e;

	if (argc != 3) {
		Tcl_AppendResult(tcl, "usage: ",
				 argv[0], "type_name {type_spec}",
				 0);
		return TCL_ERROR;
	}
	type_name = argv[1];

	strstream spec(argv[2], strlen(argv[2]));
	e = _load_type(spec, context);
	if (e != TCL_OK)
		return e;

	it = context->pop();

	/* make a new named type for the possibly anonymous type */
	/* XXX almost not implemented yet */
#if 0
	cout << "Defining '" << type_name << "' as " << *it << endl;
#endif
	/* XXX name allocation */
	context->add_type(type_name, it);

	return TCL_OK;
}

static ignore(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	tclContext *context = (tclContext *)_context;
	Type	*it;
	char	*type_name;
	int	e;

	if (argc != 2) {
		Tcl_AppendResult(tcl, "usage: ",
				 argv[0], " {type_spec}",
				 0);
		return TCL_ERROR;
	}

	strstream spec(argv[1], strlen(argv[1]));
	e = _load_type(spec, context);
	if (e != TCL_OK)
		return e;

	it = context->pop();		/* dump the value */

	if (strcmp(argv[0], "print") == 0)
		cout << *it << endl;

	return TCL_OK;
}

static collections(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	int	e;
	tclContext *context = (tclContext *)_context;
	Type	*it;
	Type	*of;

	if (argc != 2) {
		Tcl_AppendResult(tcl, "usage: ", argv[0], " {type_spec;}", 0);
		return TCL_ERROR;
	}

	strstream spec(argv[1], strlen(argv[1]));
	e = _load_type(spec, context);
	if (e != TCL_OK)
		return TCL_ERROR;

	of = context->pop();
	if (strcmp(argv[0], "bag") == 0) 
		it = new BagType(of);
	else if (strcmp(argv[0], "set") == 0)
		it = new SetType(of);
	else if (strcmp(argv[0], "list") == 0)
		it = new ListType(of);
	else if (strcmp(argv[0], "array") == 0)
		it = new ArrayType(of);
	else
		it = &TypeError;
	
	context->push(it);
	
	return TCL_OK;
}

static objects(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	int	e;
	tclContext *context = (tclContext *)_context;
	Type	*it;
	int	defining = 0;
	char	*object_name; 

	if (argc < 2  ||  argc > 3) {
		Tcl_AppendResult(tcl, "usage: ", argv[0],
				 " name [{type_spec;}]", 0);
		return TCL_ERROR;
	}
	if (argc == 3)
		defining++;

	object_name = argv[1];

	it = context->db.types.lookup(object_name);
	if (!it) {
		it = new ObjectType(object_name);
		context->add_type(object_name, it);
	}
	

	/* forward declaration, or type definition ? */
	if (defining) {
		strstream spec(argv[2], strlen(argv[2]));

		context->push(it);
		e = _load_type(spec, context);
		context->pop();
		if (e != TCL_OK) {
			/* XXX its in 2 databases !! 
			delete it;
			*/
			return e;
		}
	}
	
	return TCL_OK;
}

/*
   with ODL semantics, structs could be nameless, and 'define'
   used to give them a name.  This is allowable because
   structs don't have pointers to structs.
   Objects, however, must be named as the object can have
   a reference to its type inside itself.
 */  
static structs(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	int	e;
	tclContext *context = (tclContext *)_context;
	Type	*it;
	int	named = 0;
	char	*struct_name = 0;
	char	*struct_spec;

	if (argc < 2  ||  argc > 3) {
		Tcl_AppendResult(tcl, "usage: ", argv[0],
				 " [name] {type_spec;}", 0);
		return TCL_ERROR;
	}
	if (argc == 3) {
		named++;
		struct_spec = argv[2];
		struct_name = argv[1];
	}
	else
		struct_spec = argv[1];

	it = new StructType(struct_name);
	
	strstream spec(struct_spec, strlen(struct_spec));

	context->push(it);
	e = _load_type(spec, context);
	if (e != TCL_OK) {
		delete context->pop();
		return TCL_ERROR;
	}

	/* a named struct is a global type def, an unnamed struct is a local
		type def that returns a anonymous type */
	if (named) {
		context->add_type(struct_name, it);
		(void) context->pop();
	}
	
	return TCL_OK;
}

static attribute(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	int	e;
	tclContext *context = (tclContext *)_context;
	Type	*attribute_type;
	char	*attribute_name;
	ContainerType	*container;

	if (argc != 3) {
		Tcl_AppendResult(tcl, "usage: ", argv[0],
				 " name {type_spec;}", 0);
		return TCL_ERROR;
	}

	attribute_name = argv[1];

	container = context->top()->isContainer();
	if (!container) {
		cerr << "TOP == " << *(context->top()) << endl;
		Tcl_AppendResult(tcl, argv[0],
				 ": context isn't a container!", 0);
		return TCL_ERROR;
	}

	strstream spec(argv[2], strlen(argv[2]));

	e = _load_type(spec, context);
	if (e != TCL_OK)
		return TCL_ERROR;

	attribute_type = context->pop();
	container->add(attribute_name, attribute_type);
	
	return TCL_OK;
}

static isa(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	int	e;
	tclContext *context = (tclContext *)_context;
	char	*isa_name;
	ObjectType	*obj;
	ObjectType	*isa;

	if (argc != 2) {
		Tcl_AppendResult(tcl, "usage: ", argv[0],
				 " object_type", 0);
		return TCL_ERROR;
	}

	isa_name = argv[1];

	obj = context->top()->isObject();
	if (!obj) {
		Tcl_AppendResult(tcl, argv[0],
				 ": context isn't an object!", 0);
		return TCL_ERROR;
	}

	strstream spec(isa_name, strlen(isa_name));

	e = _load_type(spec, context);
	if (e != TCL_OK)
		return e;

	isa = context->pop()->isObject();
	if (!isa) {
		Tcl_AppendResult(tcl, argv[0], ": ",
				 argv[1], ": not an object!", 0);
		return TCL_ERROR;
	}
				 
	obj->setSuperclass(isa);
	
	return TCL_OK;
}

static extent(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	int	e;
	tclContext *context = (tclContext *)_context;
	char	*extent_name;
	ObjectType	*obj;
	Type	*extent_type;

	if (argc != 2) {
		Tcl_AppendResult(tcl, "usage: ", argv[0],
				 " extent_name", 0);
		return TCL_ERROR;
	}

	extent_name = argv[1];

	obj = context->top()->isObject();
	if (!obj) {
		Tcl_AppendResult(tcl, argv[0],
				 ": context isn't an object!", 0);
		return TCL_ERROR;
	}

	extent_type = new SetType(obj);
	context->db.extents.add(extent_name, extent_type);
	
	return TCL_OK;
}

static method(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	int	e;
	tclContext *context = (tclContext *)_context;
	Type	*it;
	char	*method_spec;

	if (argc != 2) {
		Tcl_AppendResult(tcl, "usage: ", argv[0],
				 " {method_spec;}", 0);
		return TCL_ERROR;
	}
	method_spec = argv[1];

	it = new MethodType(&TypeVoid);
	
	strstream spec(method_spec, strlen(method_spec));

	context->push(it);
	e = _load_type(spec, context);
	if (e != TCL_OK) {
		delete context->pop();
		return TCL_ERROR;
	}

	return TCL_OK;
}

static arg(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	int	e;
	tclContext *context = (tclContext *)_context;
	Type	*arg_type;
	char	*arg_name;
	MethodType	*method;

	if (argc != 3) {
		Tcl_AppendResult(tcl, "usage: ", argv[0],
				 " arg_name {type_spec;}", 0);
		return TCL_ERROR;
	}

	arg_name = argv[1];

	method = context->top()->isMethod();
	if (!method) {
// 		cerr << "TOP == " << *(context->top()) << endl;
		Tcl_AppendResult(tcl, argv[0],
				 ": context isn't a method!", 0);
		return TCL_ERROR;
	}

	strstream spec(argv[2], strlen(argv[2]));

	e = _load_type(spec, context);
	if (e != TCL_OK)
		return TCL_ERROR;

	arg_type = context->pop();
	method->addArg(arg_name, arg_type);
	
	return TCL_OK;
}

static returns(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	int	e;
	tclContext *context = (tclContext *)_context;
	Type	*type;
	MethodType	*method;

	if (argc != 2) {
		Tcl_AppendResult(tcl, "usage: ", argv[0],
				 " {type_spec;}", 0);
		return TCL_ERROR;
	}

	method = context->top()->isMethod();
	if (!method) {
		Tcl_AppendResult(tcl, argv[0],
				 ": context isn't a method!", 0);
		return TCL_ERROR;
	}

	strstream spec(argv[1], strlen(argv[1]));

	e = _load_type(spec, context);
	if (e != TCL_OK)
		return TCL_ERROR;

	type = context->pop();
	method->setReturns(type);
	
	return TCL_OK;
}

static atype(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	tclContext *context = (tclContext *)_context;
	Type *me;

	if (argc != 1) {
		Tcl_AppendResult(tcl, "usage: ",  argv[0], 0);
		return TCL_ERROR;
	}

	me = context->db.types.force_lookup(argv[0]);
	context->push(me);

	return TCL_OK;
}

static help(char *_context, Tcl_Interp *tcl, int argc, char **argv)
{
	Tcl_AppendResult(tcl, "commands: ",
			 "help ",
			 "bag ", "set ", "list ", "array ",
			 "object ", "struct ",
			 "define ", "ignore ",
			 "print ",
			 "attr ", "isa ",
			 0);
	return TCL_OK;
}

void	tclContext::add_type(char *type_name, Type *type)
{
	if (type)
		db.types.add(type_name, type);

	Tcl_CreateSCommand(this->tcl, strdup(type_name), atype, (char *) this);
}


static void tcl_init(Tcl_Interp **_tcl, tclContext *it)
{
	char *sv = (char *)it;
	Tcl_Interp *tcl;

	tcl = Tcl_CreateInterp();

	it->tcl = tcl;

	Tcl_CreateSCommand(tcl, "help", help, sv);

	Tcl_CreateSCommand(tcl, "define", define, sv);

	Tcl_CreateSCommand(tcl, "ignore", ignore, sv);
	Tcl_CreateSCommand(tcl, "print", ignore, sv);

	Tcl_CreateSCommand(tcl, "object", objects, sv);
	Tcl_CreateSCommand(tcl, "struct", structs, sv);
	Tcl_CreateSCommand(tcl, "attr", attribute, sv);
	Tcl_CreateSCommand(tcl, "attribute", attribute, sv);
	Tcl_CreateSCommand(tcl, "isa", isa, sv);
	Tcl_CreateSCommand(tcl, "extent", extent, sv);

	Tcl_CreateSCommand(tcl, "bag", collections, sv);
	Tcl_CreateSCommand(tcl, "set", collections, sv);
	Tcl_CreateSCommand(tcl, "list", collections, sv);
	Tcl_CreateSCommand(tcl, "array", collections, sv);

	Tcl_CreateSCommand(tcl, "method", method, sv);
	Tcl_CreateSCommand(tcl, "returns", returns, sv);
	Tcl_CreateSCommand(tcl, "arg", arg, sv);

	/* Force add the atomic types */
	/* ick this is disgusting -- need iterator on typedb */
	it->add_type("integer");
	it->add_type("char");
	it->add_type("boolean");
	it->add_type("real");
	it->add_type("string");
	
	*_tcl = tcl;
}

static void tcl_finish(Tcl_Interp *tcl)
{
	Tcl_DeleteInterp(tcl);
}

static	int	_load_type(istream &from, tclContext *context)
{
	return tcl_from_stream(from, cout, cerr, "> ", context->tcl, 1);
}

rc_t	load_type(istream &from, oqlDatabase &db)
{
	int	e;
	Tcl_Interp	*tcl;
	rc_t	rc;
	tclContext	context(db);

	tcl_init(&tcl, &context);
	e = _load_type(from, &context);
	tcl_finish(tcl);

	if (e == TCL_OK) {
		rc = RCOK;
	}
	else {
		rc = RC_GENERIC;
	}

	return rc;
}

#ifdef TEST_MAIN

#include "db.hh"

main()
{
	oqlDatabase	db;

	init_atomic(db);

	load_type(cin, db);

	cout << db.types << endl;
	
	exit(0);
}

#endif /* TEST_MAIN */
